@@ -39,12 +39,13 @@ import qualified Database.Persist.Sql.Util as Util
 --
 -- @since 2.17.1.0
 migrateStructured
-    :: [EntityDef]
+    :: BackendSpecificOverrides
+    -> [EntityDef]
     -> (Text -> IO Statement)
     -> EntityDef
     -> IO (Either [Text] [AlterDB])
-migrateStructured allDefs getter entity =
-    migrateEntitiesStructured getter allDefs [entity]
+migrateStructured overrides allDefs getter entity =
+    migrateEntitiesStructured overrides getter allDefs [entity]
 
 -- | Returns a structured representation of all of the DB changes required to
 -- migrate the listed entities from their current state in the database to the
@@ -54,15 +55,16 @@ migrateStructured allDefs getter entity =
 --
 -- @since 2.14.1.0
 migrateEntitiesStructured
-    :: (Text -> IO Statement)
+    :: BackendSpecificOverrides
+    -> (Text -> IO Statement)
     -> [EntityDef]
     -> [EntityDef]
     -> IO (Either [Text] [AlterDB])
-migrateEntitiesStructured getStmt allDefs defsToMigrate = do
+migrateEntitiesStructured overrides getStmt allDefs defsToMigrate = do
     r <- collectSchemaState getStmt (map getEntityDBName defsToMigrate)
     pure $ case r of
         Right schemaState ->
-            migrateEntitiesFromSchemaState schemaState allDefs defsToMigrate
+            migrateEntitiesFromSchemaState overrides schemaState allDefs defsToMigrate
         Left err ->
             Left [err]
 
@@ -73,11 +75,12 @@ migrateEntitiesStructured getStmt allDefs defsToMigrate = do
 --
 -- @since 2.17.1.0
 mockMigrateStructured
-    :: [EntityDef]
+    :: BackendSpecificOverrides
+    -> [EntityDef]
     -> EntityDef
     -> [AlterDB]
-mockMigrateStructured allDefs entity =
-    migrateEntityFromSchemaState EntityDoesNotExist allDefs entity
+mockMigrateStructured overrides allDefs entity =
+    migrateEntityFromSchemaState overrides EntityDoesNotExist allDefs entity
 
 -- | In order to ensure that generating migrations is fast and avoids N+1
 -- queries, we split it into two phases. The first phase involves querying the
@@ -532,11 +535,12 @@ mapLeft _ (Right x) = Right x
 mapLeft f (Left x) = Left (f x)
 
 migrateEntitiesFromSchemaState
-    :: SchemaState
+    :: BackendSpecificOverrides
+    -> SchemaState
     -> [EntityDef]
     -> [EntityDef]
     -> Either [Text] [AlterDB]
-migrateEntitiesFromSchemaState (SchemaState schemaStateMap) allDefs defsToMigrate =
+migrateEntitiesFromSchemaState overrides (SchemaState schemaStateMap) allDefs defsToMigrate =
     let
         go :: EntityDef -> Either Text [AlterDB]
         go entity = do
@@ -544,7 +548,7 @@ migrateEntitiesFromSchemaState (SchemaState schemaStateMap) allDefs defsToMigrat
                 name = getEntityDBName entity
             case Map.lookup name schemaStateMap of
                 Just entityState ->
-                    Right $ migrateEntityFromSchemaState entityState allDefs entity
+                    Right $ migrateEntityFromSchemaState overrides entityState allDefs entity
                 Nothing ->
                     Left $ T.pack $ "No entry for entity in schemaState: " <> show name
      in
@@ -553,11 +557,12 @@ migrateEntitiesFromSchemaState (SchemaState schemaStateMap) allDefs defsToMigrat
             (errs, _) -> Left errs
 
 migrateEntityFromSchemaState
-    :: EntitySchemaState
+    :: BackendSpecificOverrides
+    -> EntitySchemaState
     -> [EntityDef]
     -> EntityDef
     -> [AlterDB]
-migrateEntityFromSchemaState schemaState allDefs entity =
+migrateEntityFromSchemaState overrides schemaState allDefs entity =
     case schemaState of
         EntityDoesNotExist ->
             (addTable newcols entity) : uniques ++ references ++ foreignsAlt
@@ -577,7 +582,7 @@ migrateEntityFromSchemaState schemaState allDefs entity =
                 acs' ++ ats'
   where
     name = getEntityDBName entity
-    (newcols', udefs, fdefs) = postgresMkColumns allDefs entity
+    (newcols', udefs, fdefs) = postgresMkColumns overrides allDefs entity
     newcols = filter (not . safeToRemove entity . cName) newcols'
     udspair = map udToPair udefs
 
@@ -822,10 +827,13 @@ refName (EntityNameDB table) (FieldNameDB column) =
         | otherwise = shortenNames overhead (x, y - 1)
 
 postgresMkColumns
-    :: [EntityDef] -> EntityDef -> ([Column], [UniqueDef], [ForeignDef])
-postgresMkColumns allDefs t =
+    :: BackendSpecificOverrides
+    -> [EntityDef]
+    -> EntityDef
+    -> ([Column], [UniqueDef], [ForeignDef])
+postgresMkColumns overrides allDefs t =
     mkColumns allDefs t $
-        setBackendSpecificForeignKeyName refName emptyBackendSpecificOverrides
+        setBackendSpecificForeignKeyName refName overrides
 
 -- | Check if a column name is listed as the "safe to remove" in the entity
 -- list.
@@ -1097,80 +1105,87 @@ findAlters
     -> EntityDef
     -- ^ The entity definition for the entity that we're working on.
     -> Column
-    -- ^ The column that we're searching for potential alterations for.
+    -- ^ The column that we're searching for potential alterations for, derived
+    -- from the Persistent EntityDef. That is: this is how we _want_ the column
+    -- to look, and not necessarily how it actually looks in the database right
+    -- now.
     -> [Column]
+    -- ^ The columns for this table, as they currently exist in the database.
     -> ([AlterColumn], [Column])
-findAlters defs edef col@(Column name isNull sqltype def _gen _defConstraintName _maxLen ref) cols =
-    case List.find (\c -> cName c == name) cols of
+findAlters defs edef newCol oldCols =
+    case List.find (\c -> cName c == cName newCol) oldCols of
         Nothing ->
-            ([AddColumn col], cols)
+            ([AddColumn newCol] ++ refAdd (cReference newCol), oldCols)
         Just
-            (Column _oldName isNull' sqltype' def' _gen' _defConstraintName' _maxLen' ref') ->
+            oldCol ->
                 let
                     refDrop Nothing = []
                     refDrop (Just ColumnReference{crConstraintName = cname}) =
                         [DropReference cname]
 
-                    refAdd Nothing = []
-                    refAdd (Just colRef) =
-                        case find ((== crTableName colRef) . getEntityDBName) defs of
-                            Just refdef
-                                | Just _oldName /= fmap fieldDB (getEntityIdField edef) ->
-                                    [ AddReference
-                                        (crTableName colRef)
-                                        (crConstraintName colRef)
-                                        (name NEL.:| [])
-                                        (NEL.toList $ Util.dbIdColumnsEsc escapeF refdef)
-                                        (crFieldCascade colRef)
-                                    ]
-                            Just _ -> []
-                            Nothing ->
-                                error $
-                                    "could not find the entityDef for reftable["
-                                        ++ show (crTableName colRef)
-                                        ++ "]"
                     modRef =
-                        if equivalentRef ref ref'
+                        if equivalentRef (cReference oldCol) (cReference newCol)
                             then []
-                            else refDrop ref' ++ refAdd ref
-                    modNull = case (isNull, isNull') of
+                            else refDrop (cReference oldCol) ++ refAdd (cReference newCol)
+                    modNull = case (cNull newCol, cNull oldCol) of
                         (True, False) -> do
-                            guard $ Just name /= fmap fieldDB (getEntityIdField edef)
-                            pure (IsNull col)
+                            guard $ Just (cName newCol) /= fmap fieldDB (getEntityIdField edef)
+                            pure (IsNull newCol)
                         (False, True) ->
                             let
-                                up = case def of
+                                up = case cDefault newCol of
                                     Nothing -> id
-                                    Just s -> (:) (UpdateNullToValue col s)
+                                    Just s -> (:) (UpdateNullToValue newCol s)
                              in
-                                up [NotNull col]
+                                up [NotNull newCol]
                         _ -> []
                     modType
-                        | sqlTypeEq sqltype sqltype' = []
+                        | sqlTypeEq (cSqlType newCol) (cSqlType oldCol) = []
                         -- When converting from Persistent pre-2.0 databases, we
                         -- need to make sure that TIMESTAMP WITHOUT TIME ZONE is
                         -- treated as UTC.
-                        | sqltype == SqlDayTime && sqltype' == SqlOther "timestamp" =
-                            [ ChangeType col sqltype $
+                        | cSqlType newCol == SqlDayTime && cSqlType oldCol == SqlOther "timestamp" =
+                            [ ChangeType newCol (cSqlType newCol) $
                                 T.concat
                                     [ " USING "
-                                    , escapeF name
+                                    , escapeF (cName newCol)
                                     , " AT TIME ZONE 'UTC'"
                                     ]
                             ]
-                        | otherwise = [ChangeType col sqltype ""]
+                        | otherwise = [ChangeType newCol (cSqlType newCol) ""]
                     modDef =
-                        if def == def'
-                            || isJust (T.stripPrefix "nextval" =<< def')
+                        if cDefault newCol == cDefault oldCol
+                            || isJust (T.stripPrefix "nextval" =<< cDefault oldCol)
                             then []
-                            else case def of
-                                Nothing -> [NoDefault col]
-                                Just s -> [Default col s]
+                            else case cDefault newCol of
+                                Nothing -> [NoDefault newCol]
+                                Just s -> [Default newCol s]
                     dropSafe =
-                        if safeToRemove edef name
-                            then error "wtf" [Drop col (SafeToRemove True)]
+                        if safeToRemove edef (cName newCol)
+                            then error "wtf" [Drop newCol (SafeToRemove True)]
                             else []
                  in
                     ( modRef ++ modDef ++ modNull ++ modType ++ dropSafe
-                    , filter (\c -> cName c /= name) cols
+                    , filter (\c -> cName c /= cName newCol) oldCols
                     )
+  where
+    refAdd Nothing = []
+    -- This check works around a bug where persistent will sometimes
+    -- generate an erroneous ForeignRef for ID fields.
+    -- See: https://github.com/yesodweb/persistent/issues/1615
+    refAdd _ | fmap fieldDB (getEntityIdField edef) == Just (cName newCol) = []
+    refAdd (Just colRef) =
+        case find ((== crTableName colRef) . getEntityDBName) defs of
+            Just refdef ->
+                [ AddReference
+                    (crTableName colRef)
+                    (crConstraintName colRef)
+                    (cName newCol NEL.:| [])
+                    (NEL.toList $ Util.dbIdColumnsEsc escapeF refdef)
+                    (crFieldCascade colRef)
+                ]
+            Nothing ->
+                error $
+                    "could not find the entityDef for reftable["
+                        ++ show (crTableName colRef)
+                        ++ "]"
