@@ -60,6 +60,17 @@ AdminUser sql=admin_users
 
     promotedByUserId UserId
     UniquePromotedByUserId promotedByUserId
+
+FKParent sql=migration_fk_parent
+
+FKChildV1 sql=migration_fk_child
+
+-- Simulate creating a new FK field on an existing table
+FKChildV2 sql=migration_fk_child
+    parentId FKParentId
+
+ExplicitPrimaryKey sql=explicit_primary_key
+    Id Text
 |]
 
 userEntityDef :: EntityDef
@@ -77,6 +88,20 @@ password2EntityDef = entityDef (Proxy :: Proxy Password2)
 adminUserEntityDef :: EntityDef
 adminUserEntityDef = entityDef (Proxy :: Proxy AdminUser)
 
+fkParentEntityDef :: EntityDef
+fkParentEntityDef = entityDef (Proxy :: Proxy FKParent)
+
+fkChildV1EntityDef :: EntityDef
+fkChildV1EntityDef = entityDef (Proxy :: Proxy FKChildV1)
+
+fkChildV2EntityDef :: EntityDef
+fkChildV2EntityDef = entityDef (Proxy :: Proxy FKChildV2)
+
+explicitPrimaryKeyEntityDef :: EntityDef
+explicitPrimaryKeyEntityDef = entityDef (Proxy :: Proxy ExplicitPrimaryKey)
+
+-- Note that FKChild is deliberately omitted here because we have two
+-- versions of it
 allEntityDefs :: [EntityDef]
 allEntityDefs =
     [ userEntityDef
@@ -84,8 +109,11 @@ allEntityDefs =
     , passwordEntityDef
     , password2EntityDef
     , adminUserEntityDef
+    , fkParentEntityDef
+    , explicitPrimaryKeyEntityDef
     ]
 
+-- Note that this function migrates to the schema expected by FKChildV1
 migrateManually :: (HasCallStack, MonadIO m) => SqlPersistT m ()
 migrateManually = do
     cleanDB
@@ -150,6 +178,9 @@ migrateManually = do
             , "  ADD CONSTRAINT unique_promoted_by_user_id"
             , "  UNIQUE(promoted_by_user_id);"
             ]
+    rawEx "CREATE TABLE migration_fk_parent(id int8 primary key);"
+    rawEx "CREATE TABLE migration_fk_child(id int8 primary key);"
+    rawEx "CREATE TABLE explicit_primary_key(id text primary key);"
     rawEx "CREATE TABLE ignored(id int8 primary key);"
 
 cleanDB :: (HasCallStack, MonadIO m) => SqlPersistT m ()
@@ -162,6 +193,9 @@ cleanDB = do
     rawEx "DROP TABLE IF EXISTS ignored;"
     rawEx "DROP TABLE IF EXISTS admin_users;"
     rawEx "DROP TABLE IF EXISTS users;"
+    rawEx "DROP TABLE IF EXISTS migration_fk_child;"
+    rawEx "DROP TABLE IF EXISTS migration_fk_parent;"
+    rawEx "DROP TABLE IF EXISTS explicit_primary_key;"
 
 spec :: Spec
 spec = describe "MigrationSpec" $ do
@@ -551,7 +585,12 @@ spec = describe "MigrationSpec" $ do
 
         getter <- getStmtGetter
         result <-
-            liftIO $ migrateEntitiesStructured getter allEntityDefs allEntityDefs
+            liftIO $
+                migrateEntitiesStructured
+                    emptyBackendSpecificOverrides
+                    getter
+                    allEntityDefs
+                    allEntityDefs
 
         cleanDB
 
@@ -568,7 +607,12 @@ spec = describe "MigrationSpec" $ do
 
         getter <- getStmtGetter
         result <-
-            liftIO $ migrateEntitiesStructured getter allEntityDefs allEntityDefs
+            liftIO $
+                migrateEntitiesStructured
+                    emptyBackendSpecificOverrides
+                    getter
+                    allEntityDefs
+                    allEntityDefs
 
         cleanDB
 
@@ -580,5 +624,64 @@ spec = describe "MigrationSpec" $ do
             Right alters -> do
                 traverse_ (flip rawExecute [] . snd . showAlterDb) alters
                 result2 <-
-                    liftIO $ migrateEntitiesStructured getter allEntityDefs allEntityDefs
+                    liftIO $
+                        migrateEntitiesStructured
+                            emptyBackendSpecificOverrides
+                            getter
+                            allEntityDefs
+                            allEntityDefs
                 result2 `shouldBe` Right []
+
+    it "suggests FK constraints for new fields first time" $ runConnAssert $ do
+        migrateManually
+
+        getter <- getStmtGetter
+        result <-
+            liftIO $
+                migrateEntitiesStructured
+                    emptyBackendSpecificOverrides
+                    getter
+                    (fkChildV2EntityDef : allEntityDefs)
+                    [fkChildV2EntityDef]
+
+        cleanDB
+
+        case result of
+            Right [] ->
+                pure ()
+            Left err ->
+                expectationFailure $ show err
+            Right alters ->
+                map (snd . showAlterDb) alters
+                    `shouldBe` [ "ALTER TABLE \"migration_fk_child\" ADD COLUMN \"parent_id\" INT8 NOT NULL"
+                               , "ALTER TABLE \"migration_fk_child\" ADD CONSTRAINT \"migration_fk_child_parent_id_fkey\" FOREIGN KEY(\"parent_id\") REFERENCES \"migration_fk_parent\"(\"id\") ON DELETE RESTRICT  ON UPDATE RESTRICT"
+                               ]
+
+    it "Uses overrides for empty cascade action" $ runConnAssert $ do
+        migrateManually
+
+        getter <- getStmtGetter
+
+        let
+            overrideWithDefault =
+                setBackendSpecificForeignKeyCascadeDefault Cascade emptyBackendSpecificOverrides
+        result <-
+            liftIO $
+                migrateEntitiesStructured
+                    overrideWithDefault
+                    getter
+                    (fkChildV2EntityDef : allEntityDefs)
+                    [fkChildV2EntityDef]
+
+        cleanDB
+
+        case result of
+            Right [] ->
+                pure ()
+            Left err ->
+                expectationFailure $ show err
+            Right alters ->
+                map (snd . showAlterDb) alters
+                    `shouldBe` [ "ALTER TABLE \"migration_fk_child\" ADD COLUMN \"parent_id\" INT8 NOT NULL"
+                               , "ALTER TABLE \"migration_fk_child\" ADD CONSTRAINT \"migration_fk_child_parent_id_fkey\" FOREIGN KEY(\"parent_id\") REFERENCES \"migration_fk_parent\"(\"id\") ON DELETE CASCADE  ON UPDATE CASCADE"
+                               ]
